# encoding=utf-8

import warnings

from logbook import Logger
from requests import codes, HTTPError

from hansken.remote import Connection
from hansken.tool import add_command
from hansken.util import omit_empty


log = Logger(__name__)


def run_grant_command(args):
    if not args.endpoint:
        grant_parser.error('the following argument is required: ENDPOINT')
    if not args.keystore:
        grant_parser.error('the following argument is required: KEYSTORE')

    grant(args.endpoint, args.keystore,
          # we'll need either project or image, argparse's mutex group will make sure exactly one is provided
          project=args.project, image=args.image, identities=args.identities,
          auth=args.auth, verify=args.verify)


def run_grant(*args, **kwargs):
    warnings.warn('{0}.run_grant has been renamed to {0}.run_grant_command'.format(__name__),
                  DeprecationWarning)
    run_grant_command(*args, **kwargs)


def grant(base_url, keystore_url, project=None, image=None, identities=None, auth=None,
          verify=True):
    with Connection(base_url, keystore_url, auth=auth, verify=verify) as connection:
        # take either the specified image or all images in the specified project
        images = [image] if image else [linked['id'] for linked in connection.project_images(project)]
        log.debug('retrieving keys for {} images', len(images))

        keys = omit_empty({image: connection.key(image, raise_on_not_found=False) for image in images})
        log.info('{} has {} keys for {} images', connection.identity, len(keys), len(images))

        domain = connection.identity
        domain = domain[domain.index('@') + 1:]
        log.debug("determined active identity's domain to be {}", domain)

        # TODO: show progress info for total set of keys to be stored (HANSKENPY-45)
        stored_total = 0
        for identity in identities:
            if '@' not in identity:
                log.debug('expanding identity {identity} to {identity}@{domain}', identity=identity, domain=domain)
                identity = '{}@{}'.format(identity, domain)

            stored_identity = 0
            for image, key in keys.items():
                log.debug('store key for image {} for identity {}', image, identity)
                try:
                    connection.store_key(image, key, identity=identity)
                    stored_identity += 1
                except HTTPError as e:
                    # store_key will raise a 400 bad request http error if this identity already has a key for the image
                    # this is no problem here
                    if e.response is None or e.response.status_code != codes.bad_request:
                        # any other error is treated as a real error
                        raise

            log.info('stored {}/{} keys for identity {}', stored_identity, len(keys), identity)
            stored_total += stored_identity

        log.info('stored {} keys for {} identities', stored_total, len(identities))


# define command grant
grant_parser = add_command('grant',
                           run_grant_command,
                           optional_project=False,
                           help='grant keys for specific or project images to target identity or identities')
grant_parser.add_argument('identities', metavar='IDENTITY', nargs='+',
                          help='grant keys to IDENTITY')

# input takes either a single image or a single project, make argparse enforce this
images_group = grant_parser.add_mutually_exclusive_group(required=True)
images_group.add_argument('-p', '--project', metavar='PROJECT',
                          help='grant keys for all images in PROJECT')
images_group.add_argument('-i', '--image', metavar='IMAGE',
                          help='grant keys for a specific IMAGE')



# encoding=utf-8

from argparse import Action
from itertools import chain
import sys
import warnings

from logbook import Logger
from more_itertools import chunked

from hansken.query import DEFAULT_MAX_CLAUSE_COUNT, Exists, Facet, Or, Term
from hansken.remote import ProjectContext
from hansken.tool import add_command


log = Logger(__name__)


class SetItemAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        # assume we'll get two values: the key and the value to be set
        # TODO: should raise type error when this unpacking fails (HANSKENPY-40)
        key, value = values
        # get the current state of the dict we're writing to or default to a new one
        stored = getattr(namespace, self.dest, None) or {}
        # set values from the command line
        stored[key] = value
        # set the current state of the dict (possibly overwriting the old one)
        setattr(namespace, self.dest, stored)


def run_extract_command(args, progress=sys.stdout.isatty()):
    """
    Runs extractions of images in the context of a project. Requires a keystore
    to work with encrypted data; keys will be fetched automatically.

    :param args: an `argparse.Namespace` object carrying command line arguments
    :param progress: whether to include progress info to stdout (defaults to
        ``sys.stdout.isatty()``)
    """
    # check that we'll have an endpoint and a project (error message emulated to look like argparse's own messages)
    if not args.endpoint:
        extract_parser.error('the following argument is required: ENDPOINT')
    if not args.project:
        extract_parser.error('the following argument is required: PROJECT')
    if not args.keystore:
        extract_parser.error('the following argument is required: KEYSTORE')

    def filter_images():
        # find current states of the images
        filtered = set()

        for image in linked:
            if image['id'] in images:
                # image is to be processed, find its current state
                image['state'] = image['states'][-1]['state']
                properties = image.items()
                # match all key:value pairs from the filter
                # TODO: enable negative filters (HANSKENPY-40)
                if all(prop in properties for prop in args.filter.items()):
                    log.debug('including image {}', image['id'])
                    filtered.add(image['id'])
                else:
                    log.debug('excluding image {}', image['id'])

        return filtered

    def images_with_user_content():
        # query to find user-generated content (notes and tags) within the images
        user_query = Exists('#note') | Exists('tags')
        # facet to get counters for images that have user-generated content
        facet = Facet('image')

        log.debug('checking for user content in {} images', len(images))

        def user_content_images(batch):
            query = user_query & Or(*[Term('image', image) for image in batch])
            # the keys with hits in the facets will be image id's that match user_query
            return context.search(query, count=0, facets=facet).facets[0].keys()

        return set(chain.from_iterable(
            # generate images that contain user-generated content in batches close to the max clause count
            user_content_images(batch) for batch in chunked(images, DEFAULT_MAX_CLAUSE_COUNT - 4)
        ))

    def what_to_extract():
        # provide information on the images that have user content
        print('the following images have user content within project {}:'.format(args.project))
        infos = [image for image in linked if image['id'] in user_content]
        for image in infos:
            print('  {id} ({description})'.format(**image))

        # ask the user what to do with that information (include / exclude these images or quit)
        try:
            include = None
            while include is None:
                # get either an empty string (default) or the first lowered char of whatever was entered
                choice = input('extract these images (y/[n]/q)? ')[:1].lower()
                if choice in 'ynq':  # NB: '' in 'ynq' is True
                    if choice == 'q':
                        raise SystemExit()
                    else:
                        # only 'yessy' should mean include (empty string = default = no)
                        include = choice == 'y'
        except (EOFError, KeyboardInterrupt):
            log.debug('user content inclusion choice cancelled by user')
            # user pressed ^D or ^C, abort command
            raise SystemExit('aborted')

        # return the chosen set of images to be extracted
        return images if include else images - user_content

    with ProjectContext(args.endpoint, args.project, args.keystore,
                        auth=args.auth, verify=args.verify) as context:
        linked = context.images()

        if args.images:
            # verify all images to be extracted are linked to project
            not_linked = set(args.images) - {image['id'] for image in linked}
            if not_linked:
                raise ValueError('images not linked to project {}: {}'.format(
                    args.project, ', '.join(not_linked)
                ))

        # create base set of images, either read from command line or defaulting to all linked images
        images = set(args.images) or {image['id'] for image in linked}

        if args.filter:
            log.info('filtering {} images using {} filters', len(images), len(args.filter))
            images = filter_images()
            log.info('{} images left after applying filters', len(images))

        if args.skip_images:
            skip_images = set(args.skip_images)
            log.info('skipping {} images', len(skip_images))
            images = images - skip_images
            log.info('{} images left after removing skipped images', len(images))

        if images and args.check_user_content:
            user_content = images_with_user_content()
            if user_content:
                log.info('{} images have user content, prompting user what to do', len(user_content))
                images = what_to_extract()
                log.info('{} images left after user content check', len(images))

        tools = None
        if args.include_tools or args.exclude_tools:
            include = set(args.include_tools or [])
            exclude = set(args.exclude_tools or [])

            if include & exclude:
                raise ValueError('cannot both include and exclude tools [{}]'.format(
                    ', '.join(sorted(include & exclude)))
                )

            # select all default tools
            tools = {name for name, tool in context.connection.tools().items() if tool.get('defaultEnabled')}
            log.info('default tool set uses {} tools', len(tools))

            # add and subtract tools to be included or excluded
            tools |= include
            tools -= exclude
            log.info('using {} tools after including [{}] and excluding [{}]',
                     len(tools),
                     ', '.join(sorted(include)),
                     ', '.join(sorted(exclude)))

        # make sure tool names becomes a list (sorted to be deterministic) or remains None to select the default
        tools = sorted(tools) if tools else None

        # start extractions
        if images:
            log.info('extracting {} images as user {}', len(images), context.connection.identity)
            try:
                if progress:
                    print('\033[2K\r[{:<40}] ({}/{})'.format(
                        '',  # no progress yet
                        0, len(images)
                    ), end='')
                else:
                    # provide some output if we won't be using a fancy progress bar
                    print('extracting {} images...'.format(len(images)))

                num_errors = 0
                for num, image in enumerate(images, start=1):
                    try:
                        # the actual action we set out to do (key left at automatic)
                        context.connection.extract(args.project, image, tools=tools, configuration=args.config)
                    except Exception as e:
                        num_errors += 1
                        log.exception('extraction of image {} within project {} failed', image, args.project, e)
                    finally:
                        if progress:
                            # print a 40-char progress bar with number indicator
                            print('\033[2K\r[{:<40}] ({}/{}){}'.format(
                                # fill progress bar as far as needed
                                '»' * int(40 * num / len(images)),
                                # numeric progress info
                                num, len(images),
                                # append number of errors if > 0
                                ' ({} errors)'.format(num_errors) if num_errors > 0 else ''
                            ), end='')

                # set return code to 1 on any number of errors
                return min(1, num_errors)
            finally:
                if progress:
                    # append a newline after our progress bar
                    print()
        else:
            raise SystemExit('no images to extract')


def run_extract(*args, **kwargs):
    warnings.warn('{0}.run_extract has been renamed to {0}.run_extract_command'.format(__name__),
                  DeprecationWarning)
    return run_extract_command(*args, **kwargs)


extract_parser = add_command('extract', run_extract_command, optional_project=False,
                             help='start extraction of images linked to a project')
extract_parser.add_argument('project', metavar='PROJECT',
                            help='the project to extract images for')
extract_parser.add_argument('images', metavar='IMAGE', nargs='*',
                            help='the images to be extracted (defaults to all images linked to PROJECT)')

extract_parser.add_argument('-s', '--state', metavar='STATE', dest='filter',
                            # 'parse' the passed value for state into the expected two-tuple
                            type=lambda value: ('state', value), action=SetItemAction,
                            help='consider only images whose current state is STATE')
extract_parser.add_argument('--property', metavar='PROPERTY:VALUE', dest='filter',
                            # parse the value passed value by splitting on :
                            # TODO: should raise TypeError on 'syntax error' (HANSKENPY-40) (or should use nargs=2)
                            type=lambda value: value.split(':', 1), action=SetItemAction,
                            help='consider only images whose value for PROPERTY is VALUE')
extract_parser.add_argument('-c', '--check-user-content', dest='check_user_content', action='store_true', default=True,
                            help='check if any selected images have user content and prompt for action to take '
                                 '(this is the default)')
extract_parser.add_argument('-C', '--no-check-user-content', dest='check_user_content', action='store_false',
                            help="don't check for user content")
extract_parser.add_argument('--skip-images', metavar='IMAGE', nargs='+',
                            help='images to be skipped, even if they would be included by other filters')
extract_parser.add_argument('--config', metavar='KEY:VALUE', dest='config',
                            # TODO: should raise TypeError on 'syntax error' (HANSKENPY-40) (or should use nargs=2)
                            type=lambda value: value.split(':', 1), action=SetItemAction,
                            help='override extraction configuration key CONFIG with VALUE for this single extraction')
extract_parser.add_argument('--include-tools', metavar='TOOL', nargs='+',
                            help='include named TOOL with the tools to be used to extract IMAGE(S)')
extract_parser.add_argument('--exclude-tools', nargs='+',
                            help='exclude named TOOL from the tools to be used to extract IMAGE(S)')



# encoding=utf-8

from collections.abc import Iterable, Sized
from datetime import date, datetime
import json
import re
import warnings

from logbook import Logger

from hansken.util import format_datetime, GeographicLocation, omit_empty, Vector


log = Logger(__name__)


# 'known' default maximum number of clauses in a (boolean) query
DEFAULT_MAX_CLAUSE_COUNT = 1024


# collect the type of a compiled pattern (re.Pattern in 3.7+, a runtime type in Python 3.6)
_Pattern = type(re.compile('.*'))


def _format_value(value):
    """
    Formats a value suitable for Hansken Query Language:

    - date, datetime: ISO 8601
    - GeographicLocation: ISO 6709 string
    - tuple of numbers, length 2: assumed to be latlong, formatted as an
      ISO 6709 string.
    - others: no change
    """
    # order is significant here, date is a datetime but date is allowed without a timezone
    if isinstance(value, datetime):
        return format_datetime(value)
    if isinstance(value, date):
        return value.isoformat()

    if isinstance(value, GeographicLocation):
        return str(value)
    if isinstance(value, (tuple, list)) and len(value) == 2 and all(isinstance(part, (float, int)) for part in value):
        return str(GeographicLocation(*value))

    return value


def _flatten(q_type, *queries):
    """
    Yields queries or contained clauses if a query is of type q_type.

    :param q_type: (iterable) type of query to flatten
    :param queries: queries to process
    :return: generator yielding queries
    """
    for q in queries:
        if isinstance(q, q_type):
            # flatten the top level only
            for clause in q:
                yield clause
        else:
            yield q


def _parse_scale(scale):
    """
    Parses a scale string into a scale and base or interval as defined by the
    Facet type of the Hansken search request. Parses a scale such as 'log2',
    'log10' or 'linear@1024'.

    :param scale: a scale string
    :return: (scale, arg)
    """
    base = None
    interval = None
    if scale and scale.startswith('log') and len(scale) > len('log'):
        base = int(scale[len('log'):])
        scale = 'log'
    elif scale and scale.startswith('linear@'):
        interval = int(scale[len('linear@'):])
        scale = 'linear'

    return scale, base, interval


class Query:
    """
    Base class for Hansken query types. Implementations are required to
    implement `.as_dict` for transformation to wire format.
    """

    def as_dict(self):
        """
        Turns this query into a dict as specified by the Hansken Query
        Language Specification.
        """
        raise NotImplementedError()

    def __and__(self, other):
        """
        Binary and operator (``&``) handling, resulting in an `.And` query.
        Resulting query is flattened when one or more operands are already
        `.And` queries.
        """
        if not isinstance(other, Query):
            raise TypeError('right hand operand not Query')

        return And(*_flatten(And, self, other))

    def __or__(self, other):
        """
        Binary or operator (``|``) handling, resulting in an `.Or` query.
        Resulting query is flattened when one or more operands are already
        `.Or` queries.
        """
        if not isinstance(other, Query):
            raise TypeError('right hand operand not Query')

        return Or(*_flatten(Or, self, other))

    def __invert__(self):
        """
        Binary not operator (``~``) handling, resulting in a `.Not` query.
        """
        return Not(self)

    def __str__(self):
        """
        Encodes this query as a JSON string.
        """
        return json.dumps(self.as_dict())


class And(Sized, Iterable, Query):
    """
    Boolean conjunction of multiple queries; traces should match all
    contained queries, for example:

    .. code-block:: python

        And(Term('file.name', 'query.py'),
            Range('data.raw.size', min=512))
    """

    def __init__(self, *queries):
        self.queries = set(queries)

    def as_dict(self):
        return {'and': [query.as_dict() for query in self.queries]}

    def __len__(self):
        return len(self.queries)

    def __iter__(self):
        return iter(self.queries)


class Or(Sized, Iterable, Query):
    """
    Boolean disjunction of multiple queries, traces should match any
    contained query, for example:

    .. code-block:: python

        Or(Term('file.name', 'query.py'),
           Range('data.raw.size', max=1024))
    """

    def __init__(self, *queries):
        self.queries = set(queries)

    def as_dict(self):
        return {'or': [query.as_dict() for query in self.queries]}

    def __len__(self):
        return len(self.queries)

    def __iter__(self):
        return iter(self.queries)


class Not(Query):
    """
    Negates a single query, for example:

    .. code-block:: python

        Not(Term('file.name', 'query.py'))
    """

    def __init__(self, query):
        self.query = query

    def as_dict(self):
        return {'not': self.query.as_dict()}


class Nested(Query):
    """
    Query a field for values matching the results of another query, for example:

    .. code-block:: python

        Nested('data.raw.hash.md5', Term('file.name', 'query.py'))
    """

    def __init__(self, field, query):
        self.field = field
        self.query = query

    def as_dict(self):
        return {'nested': {
            'field': self.field,
            'query': self.query.as_dict()
        }}


class Tracelet(Query):
    """
    Restrict a query for a tracelet type to the same tracelet instance of that
    tracelet type.

    .. code-block:: python

        # find traces containing an entity
        Tracelet('entity')

        # find traces containing an entity that has both:
        # - a value starting with "http://"
        # - a confidence of at least 0.9
        Tracelet('entity', Term('entity.value', 'http://*', full=True) & Range('entity.confidence', min=0.9))

    Note that without the `Tracelet` query, the `Term` and `Range` queries above
    could match different entities, ultimately matching traces that contain
    *any* entity with a value starting with ``http://`` and *any* entity with a
    confidence of at least 0.9 (not necessarily to the same entity).
    """
    def __init__(self, tracelet_type, query=None):
        self.tracelet_type = tracelet_type
        self.query = query

    def as_dict(self):
        return {'hasTracelet': {
            'type': self.tracelet_type,
            # default to an empty dict, translating to an any-query (match any tracelet instance)
            'query': self.query.as_dict() if self.query else {}
        }}


class Trace(Query):
    """
    Restrict a tracelet query to tracelets belonging to traces matching the
    inner query.

    .. code-block:: python

        # match entities of type iban, but only if the trace they belong to is from a specific image
        Term('entity.type', 'iban') & Trace(Term('image', '1234-abcd'))
    """
    def __init__(self, query):
        self.query = query

    def as_dict(self):
        return {'hasTrace': {
            'query': self.query.as_dict()
        }}


class Term(Query):
    """
    Query for the value of single field, for example:

    .. code-block:: python

        # search for files with name "query.py"
        Term('file.name', 'query.py')
        # search for occurrences of the term "query" (in either data or metadata)
        Term('query')
    """

    def __init__(self, field_or_value, value=None, full=False):
        """
        Create a new `.Term` query.

        :param field_or_value: the field to search, or (when *value* is not
            supplied) the search value
        :param value: value to search for (only needed when searching a
            specific field)
        :param full: search the untokenized variant of any string, see
            :ref:`full matches <full_match>`
        """
        # allow first param to be the value, set field to 'text' to request a general term search
        if value is None:
            value = field_or_value
            # full matches won't work on text (which expands to both meta and data)
            field = 'meta' if full else 'text'
        else:
            field = field_or_value

        if isinstance(value, float):
            raise TypeError('float value for Term not supported: {}'.format(value))

        self.field = field
        self.value = value
        self.full = full

    def as_dict(self):
        return {'term': {
            'field': self.field,
            'value': _format_value(self.value),
            'fullMatch': self.full,
        }}


class Regex(Query):
    """
    Query a field for string-values matching a regular expression, for example:

    .. code-block:: python

        # search for replies or forwards
        Regex('email.subject', '(re|fw): .*', full=True)
        # search for bombs, or some curious misspellings
        Regex(re.compile(r'bo[mn]+bs'))

    Either a `str` or a `re.Pattern` object is accepted, of which only the
    ``pattern`` property is used.

    .. note::

        - Regular expressions always match entire terms or (in case of
          ``full=True``) properties, as if the regular expression was anchored
          at both ends, see :ref:`full matches <full_match>`.
        - Not every feature supported by Python's `re` module (like particular
          character classes (``\\s`` / ``\\w``), start/end anchors
          (``^``/``$``), look ahead/behind or non-greedy quantifiers (``??`` /
          ``*?``)) will be supported by Hansken. The use of these is *not*
          validated by ``hansken.py``, but will result in errors when submitted.
        - Regular expressions queries are *always* case insensitive and ignore
          diacritics in values.
    """

    def __init__(self, field_or_pattern, pattern=None, full=False):
        """
        Create a new `.Regex` query.

        :param field_or_pattern: the field to match, or (when *pattern* is not
            supplied) the search pattern
        :param pattern: pattern to match, either a `str` or `re.Pattern`
        :param full: match the untokenized variant of the value, see
            :ref:`full matches <full_match>`
        """
        # allow first param to be the pattern, set field to 'text' to request a general term match
        if pattern is None:
            pattern = field_or_pattern
            # full matches won't work on text (which expands to both meta and data)
            field = 'meta' if full else 'text'
        else:
            field = field_or_pattern

        if isinstance(pattern, _Pattern):
            if pattern.flags & ~re.UNICODE:
                log.warn('regular expression flags in Query objects are ignored ({} used by pattern)',
                         str(re.RegexFlag(pattern.flags)))
            # remote has no use for a Pattern object, use value's original pattern string
            pattern = pattern.pattern

        if not isinstance(pattern, str):
            # the only sanity check we can do on the client side
            raise TypeError(f'query value for Regex must be str, not {type(pattern).__name__}')

        self.field = field
        self.pattern = pattern
        self.full = full

    def as_dict(self):
        return {'regex': {
            'field': self.field,
            'value': self.pattern,
            'fullMatch': self.full,
        }}


class Range(Query):
    """
    Query a field for values in a particular range, for example:

    .. code-block:: python

        # search for traces with entropy between 4.0 (exclusive) and 7.0 (inclusive)
        Range('data.raw.entropy', gt=4.0, max=7)
        # search for traces no larger than 1MiB (1 << 20 == 2 ** 20 == 1048576 bytes)
        Range('data.raw.size', max=1 << 20)
        # search for traces with peculiar names (matches file name aab.txt, but not ccb.txt)
        Range('file.name', min='aa', max='cc')
    """
    _range_keys = {
        'gt': '>',
        'gte': '>=',
        'lt': '<',
        'lte': '<=',
        'max': '<=',
        'maxvalue': '<=',
        'max_value': '<=',
        'min': '>=',
        'minvalue': '>=',
        'min_value': '>=',
    }

    def __init__(self, field, **ranges):
        """
        Create a new `.Range` query.

        :param field: the field to query for
        :param ranges: keyword arguments of the following forms:

            - ``>``, ``gt``: value should be greater than supplied value;
            - ``>=``, ``gte``, ``min``, ``minvalue``, ``min_value``: value should
              be greater or equal to supplied value;
            - ``<``, ``lt``: value should be less than supplied value;
            - ``<=``, ``lte``, ``max``, ``maxvalue``, ``max_value``: value should
              be less or equal to supplied value;
        """
        self.field = field
        # translate kwarg type range to hansken keys
        self.ranges = {self._range_keys.get(key, key): value for key, value in ranges.items()}

        if len(ranges) != len(self.ranges):
            raise ValueError('duplicate mapping in ranges for Range')

    def as_dict(self):
        q = {'field': self.field}
        q.update(self.ranges)

        return {'range': q}


class Exists(Query):
    """
    Search for traces that have a particular field, for example:

    .. code-block:: python

        Exists('email.headers.In-Reply-To')
    """

    def __init__(self, field):
        self.field = field

    def as_dict(self):
        return {'exists': self.field}


class Phrase(Query):
    """
    Search for a phrase of terms, occurring within a particular distance of
    each other, for example:

    .. code-block:: python

        Phrase('email.subject', 'sell you a bomb')
        # will also match "sell you a bomb", not restricted to just email.subject
        Phrase('sell bomb', distance=2)
    """

    def __init__(self, field_or_value, value=None, distance=0):
        """
        Create a new `.Phrase` query.

        :param field_or_value: the field to search, or (when *value* is not
            supplied) the search value
        :param value: value to search for (only needed when searching a
            specific field)
        :param distance: the max number of position displacements between
            terms in the phrase (0 being an exact phrase match)
        """
        # allow first param to be the value, set field to 'text' to request a general phrase search
        if value is None:
            value = field_or_value
            # switch to magical property 'text', expanding to both 'meta' and 'data'
            field = 'text'
        else:
            field = field_or_value

        if len(value) > 1000:
            raise ValueError('value too long for Phrase: {}'.format(len(value)))
        if '?' in value or '*' in value:
            raise ValueError('wildcard in value not allowed for Phrase: {}'.format(value))

        self.field = field
        self.value = value
        self.distance = distance

    def as_dict(self):
        return {'phrase': {
            'field': self.field,
            'value': self.value,
            'distance': self.distance,
        }}


class GeoBox(Query):
    """
    Search for traces with location data within the bounding box between two
    corner points: southwest and northeast, for example:

    .. code-block:: python

        # a location can either be a 2-tuple (…)
        GeoBox('gps.latlong', (-1, -2), (3, 4))
        # (…) or an ISO 6709 latlong string
        GeoBox('gps.latlong', '+12.5281-070.0229', '+13.5281-080.0229')
    """

    def __init__(self, field, southwest, northeast):
        self.field = field
        self.sw = southwest
        self.ne = northeast

    def as_dict(self):
        return {'geobox': {
            'field': self.field,
            'southwest': _format_value(self.sw),
            'northeast': _format_value(self.ne),
        }}


class HQLHuman(Query):
    """
    Search for traces using HQL Human query syntax, for example:

    .. code-block:: python

        HQLHuman('file.name:query.py')
        HQLHuman('data.raw.size>1024')
    """

    def __init__(self, query):
        if not isinstance(query, str):
            raise TypeError('HQL-Human query not a string: {}'.format(query))

        self.query = query

    def as_dict(self):
        return {'human': self.query}


def to_query(query):
    """
    Make sure *query* is a `.Query` instance by either wrapping it with a
    `.HQLHuman` or returning it as is.

    :param query: either a `str` or a `.Query`
    :return: a `.Query` instance
    :raise TypeError: when *query*'s type is not acceptable
    """
    if query is None:
        return None
    if isinstance(query, str):
        return HQLHuman(query)
    if isinstance(query, Query):
        return query

    raise TypeError('query should be either str or hansken.query.Query, not {}'.format(type(query)))


_sort_direction = {
    'ascending': 'ascending',
    'asc': 'ascending',
    '+': 'ascending',
    'descending': 'descending',
    'desc': 'descending',
    '-': 'descending',
}


class Sort:
    def __init__(self, field, direction=None, filter=None, mode=None, value=None):
        """
        Creates a sort clause for use with a search request.

        The *mode* parameter determines what kind of sorting should be applied:

        - ``value``: a regular sort-by-value (the default applied by the
          remote);
        - ``exists``: simply sort on whether the sort field as a value;
        - ``cosineSimilarity``: use parameter *value* to sort on the cosine
          similarity between *value* and the value of the sort field;
        - ``manhattanDistance``: similarly sort on the manhattan distance (or
          L1 norm) between *value* and the value of the sort field;
        - ``euclideanDistance``: similarly sort on the euclidean distance (or
          L2 norm) between *value* and the value of the sort field;

        :param field: the field to sort on
        :param direction: the sorting direction (ascending or descending, or
            ``None`` to auto-determine the sort direction from other arguments)
        :param filter: an optional query to restrict the tracelets included for
                       sorting
        :param mode: a sort mode (see above)
        :param value: a (vector) value to use for similarity / distance
                      calculations in applicable sorting modes (see above)
        """
        self.field = field
        self.filter = filter
        self.mode = mode
        self.value = value

        if isinstance(self.value, Vector):
            # if value is a supplied as a Vector object, coerce it to a str
            # otherwise, leave it as (already a str, None, list of floats, …)
            self.value = str(value)

        if self.value and not self.mode:
            # mode is required when value is supplied
            # default to cosine similarity if user supplied no mode
            self.mode = 'cosineSimilarity'

        if direction:
            # require arg to be known when provided
            self.direction = _sort_direction[direction.lower()]
        else:
            # no explicit direction specified, default to ascending, unless the mode is cosineSimilarity
            # (high similarity score is 'low distance', inverting the intuitive sorting order)
            self.direction = 'descending' if self.mode == 'cosineSimilarity' else 'ascending'

    def as_dict(self):
        return omit_empty({
            'field': self.field,
            'direction': self.direction,
            'filter': to_query(self.filter).as_dict() if self.filter else None,
            'mode': self.mode,
            'value': self.value,
        })

    @classmethod
    def from_str(cls, sort):
        """
        Creates a `.Sort` from *sort*, parsing field, direction and filter.

        Formats supported:

        - ``some.field``: sort on field "some.field", ascending
        - ``some.field+``: sort on field "some.field", ascending
        - ``some.field-``: sort on field "some.field", descending
        - ``some.field | query*``: sort on field "some.field" within matches
          for query "query*", ascending (sorting non-matches after matches)

        :param sort: a sorting string to parse
        :return: a `.Sort` instance
        """
        parts = sort.split('|', 1)

        direction = None
        field = parts[0].strip()
        if field[-1] in ('+', '-'):
            direction = _sort_direction[field[-1]]
            field = field[:-1]

        filter = None
        if len(parts) == 2:
            filter = HQLHuman(parts[1].strip())

        return cls(field, direction, filter)


def to_sort(sort):
    if sort is None:
        return None
    if isinstance(sort, str):
        return Sort.from_str(sort)
    if isinstance(sort, Sort):
        return sort

    raise TypeError('sort should be either str or hansken.query.Sort, not {}'.format(type(sort)))


class Facet:
    def __init__(self, field, size=100, include_total=None, scale=None, filter=None):
        self.field = field
        self.size = size
        self.include_total = include_total
        self.filter = to_query(filter)

        if scale:
            # discourage use of direct Facet with custom parsed scale
            warnings.warn(DeprecationWarning('using scale on Facet is deprecated, use one of the sub types instead'))

        self.scale, self.base, self.interval = _parse_scale(scale)

        self.precision = self.min = self.max = self.sw = self.ne = None

    def as_dict(self):
        """
        Turns this facet into a dict as specified by the Hansken Query
        Language Specification.
        """
        return omit_empty({
            'field': self.field,
            'size': self.size,
            'includeTotal': self.include_total,
            'scale': self.scale,
            'base': self.base,
            'interval': self.interval,
            # make sure to format dates and date(time)s as ISO 8601
            'min': _format_value(self.min),
            'max': _format_value(self.max),
            'precision': self.precision,
            # make sure to format coordinates as ISO 6709
            'southwest': _format_value(self.sw),
            'northeast': _format_value(self.ne),
            'filter': self.filter.as_dict() if isinstance(self.filter, Query) else self.filter
        })

    def __str__(self):
        """
        Encodes this facet as a JSON string.
        """
        return json.dumps(self.as_dict())


class TermFacet(Facet):
    def __init__(self, field, size=100, include_total=None, filter=None):
        """
        Create a new `.TermFacet` to use with a query. A term facet can be
        created on any type of field, counting the occurrences of any value.

        :param field: field to create a facet on
        :param size: the max number of facet counters to return, default is
            100
        :param filter: only count traces matching filter
        """
        super().__init__(field=field, size=size, include_total=include_total, filter=filter)


class RangeFacet(Facet):
    def __init__(self, field, scale, base=None, interval=None, min=None, max=None, include_total=None, filter=None):
        """
        Create a new `.RangeFacet` to use with a query. A range facet can be
        made on either numeric or date fields.

        :param field: field to create a facet on
        :param scale:
            - ``year``, ``month``, ``day``, ``hour``, ``minute`` or ``second``
              for date fields
            - ``linear`` or ``log`` for numeric fields
        :param base: logarithmic base when scale is `'log'`
        :param interval: interval or bucket size when scale is `'linear'`
        :param min: minimum value to include in the facet result
        :param max: maximum value to include in the facet result
        :param filter: only count traces matching filter
        """
        if scale not in ('log', 'linear', 'year', 'month', 'day', 'hour', 'minute', 'second'):
            raise ValueError('unknown scale: {}'.format(scale))

        if scale in ('year', 'month', 'day', 'hour', 'minute', 'second') and (base or interval):
            # date range facet, base and interval are illegal
            raise ValueError('date range facet cannot be combined with base or interval')
        elif scale == 'log' and (not base or interval):
            # numeric log facet, requires base, interval is illegal
            raise ValueError('numeric log facet needs base, cannot be combined with interval')
        elif scale == 'linear' and base:
            # numeric linear facet, base is illegal
            raise ValueError('numeric linear facet cannot be combined with base')

        super().__init__(field=field, include_total=include_total, filter=filter)

        self.scale = scale
        self.base = base
        self.interval = interval
        self.min = min
        self.max = max


class GeohashFacet(Facet):
    def __init__(self, field, size=100, include_total=None, precision=1, southwest=None, northeast=None, filter=None):
        """
        Create a new `.Facet` to use with a query.

        :param field: field to create a facet on
        :param size: the max number of facet counters to return, default is
            100
        :param precision: number of characters of the returned geohashes
        :param southwest: south west bound / corner point
        :param northeast: north west bound / corner point
        :param filter: only count traces matching filter
        """
        bounds = (southwest, northeast)
        if any(bounds) and not all(bounds):
            raise ValueError('specify either both southwest and northeast bounds or neither')

        super().__init__(field=field, size=size, include_total=include_total, filter=filter)

        self.precision = precision
        self.sw = southwest
        self.ne = northeast


def to_facet(facet):
    if facet is None:
        return None
    if isinstance(facet, str):
        return TermFacet(facet)
    if isinstance(facet, Facet):
        return facet

    raise TypeError('facet should be either str or hansken.query.Facet, not {}'.format(type(facet)))



# encoding=utf-8

from argparse import SUPPRESS
from collections import deque
from datetime import datetime, timezone
from getpass import getpass
from os import path, urandom
from subprocess import PIPE, Popen, STDOUT  # nosec
import sys
from uuid import uuid4
import warnings

from logbook import Logger
from requests import HTTPError

from hansken.remote import Connection
from hansken.tool import add_command
from hansken.tool._webhdfs import WebHDFSServer
from hansken.util import b64decode, b64encode, format_byte_size, Namespace, parse_byte_size


log = Logger(__name__)


upload_info = Namespace(marker='\f',  # marker used by the image tool to indicate line of machine-readable output
                        names={
                            # name of the local file being processed
                            'inputName': ('in_name', str),
                            # image size of the input file(set)
                            'inputSize': ('in_size', int),
                            # amount of bytes read
                            'currentOffset': ('num_read', int),
                            # image id of output image
                            'outputName': ('out_name', str),
                            # amount of bytes written
                            'outputSize': ('out_size', int),
                            # final status of conversion (split into (result, status), where result can either be 'OK'
                            # or 'FAIL', status is a string-representation of the conversion status)
                            'status': ('status', lambda value: value.split(':', 1)),
                            # number of bytes converted before hitting error status
                            'bytesConverted': ('bytes_converted', int),
                        })


class local_proxy:  # noqa: N801
    def __repr__(self):
        return '<authenticated proxy>'


# overwrite authenticated_proxy with an instance of itself
local_proxy = local_proxy()


def run_upload_command(args, progress=sys.stdout.isatty()):
    """
    Prepares arguments from the command line before being passed to upload.

    :param args: an argparse.Namespace object
    :param progress: whether to print progress information during upload
                     (defaults to the result of sys.stdout.isatty())
    :return: return code (zero for success, non-zero for error)
    """
    # check that we'll have an endpoint and a project (error message emulated to look like argparse's own messages)
    if not args.endpoint:
        upload_parser.error('the following argument is required: ENDPOINT')
    if not args.project and (args.link or args.extract):
        upload_parser.error('the following argument is required (linking or extraction requested): PROJECT')
    if not args.link and args.extract:
        upload_parser.error('starting extraction impossible without linking image to project')
    if not args.key:
        upload_parser.error('the following argument is required: KEY')
    if args.store_key and not args.keystore:
        upload_parser.error('no keystore endpoint available to store key')

    def print_values():
        try:
            print('---')
            while True:
                name, value = yield  # catch 2-tuple sent to print
                print('{name:<20} {value}'.format(name=name + ':', value=value))
                sys.stdout.flush()
        finally:
            print('...')
            sys.stdout.flush()

    return_code = 0
    for image in args.images:
        key = args.key
        if callable(key):
            # generate / prompt a new key for every image being processed
            key = key()

        input_size = None
        report = print_values()
        next(report)  # initialize the report generator, pause at the yield
        report.send(('file', image))
        report.send(('started', datetime.now(tz=timezone.utc).isoformat()))
        try:
            # upload is a generator, should come up with our image id while it does its thing
            for name, value in upload(args.endpoint,
                                      # use provided image_id or generate random image id
                                      args.image_id or str(uuid4()),
                                      image,
                                      b64decode(key.strip(), validate=True),
                                      # default to local_proxy (hdfs_url is a hidden option)
                                      args.hdfs_url or local_proxy,
                                      bufsize=args.bufsize,
                                      max_retries=args.retry if args.retry >= 0 else None,
                                      retry_wait=args.wait,
                                      auth=args.auth,
                                      # only supply a project when linking
                                      project=args.project if args.link else args.link,
                                      extract=args.extract,
                                      keystore_url=args.keystore,
                                      store_key=args.store_key,
                                      verify=args.verify,
                                      image_tool_path=args.image_tool_path,
                                      streaming=args.streaming):

                if name == 'in_size':
                    report.send(('input size', format_byte_size(value)))
                    input_size = value

                if name == 'out_name':
                    report.send(('id', value))
                    report.send(('key', key))

                if progress and input_size and name == 'num_read':
                    # display progress only when input size is known
                    # clear line (\033[2K), move caret to start of line (\r), print progress line
                    print('\033[2K\r[{:<40}] ({})'.format(
                        '»' * int(40 * value / input_size),
                        format_byte_size(value)
                    ), end='')

                if name == 'out_size':
                    if progress and input_size:
                        # clear the line we've been using for progress info (hides the bar when completed)
                        print('\033[2K\r', end='')

                    report.send(('output size', format_byte_size(value)))
                    if input_size and value:
                        # if we know the input size and the output size is not 0, show the compression ratio
                        report.send(('compression ratio', round(input_size / value, 3)))

                if name == 'status':
                    report.send(('status', '{} ({})'.format(*value)))

                if name == 'bytes_converted':
                    report.send(('bytes converted', format_byte_size(value)))

            report.send(('completed', datetime.now(tz=timezone.utc).isoformat()))
        except AssertionError as e:
            return_code = 1
            print(str(e), file=sys.stderr)
            sys.stderr.flush()
        finally:
            report.close()

    return return_code


def run_upload(*args, **kwargs):
    warnings.warn('{0}.run_upload has been renamed to {0}.run_upload_command'.format(__name__),
                  DeprecationWarning)
    run_upload_command(*args, **kwargs)


def upload(connection_or_endpoint, image_id, image_file, key,
           destination=local_proxy, bufsize=8 << 20, max_retries=8, retry_wait=10.0,
           auth=None, project=False, extract=False, keystore_url=None, store_key=False,
           verify=True, image_tool_path='hansken-image-tool', streaming=False,
           **image_meta):
    """
    Calls hansken-image-tool to convert a local image file to an NFI image and
    uploads it to webhdfs (destination should include directory on target
    HDFS). Links the uploaded image to a project and starts extraction if
    requested. User and description for the created image metadata are based
    on the supplied user and image file name.

    :param connection_or_endpoint: either an instance of `.Connection` or an
        endpoint to which a new `.Connection` instance will be connected
    :param image_id: image id of the image to be uploaded
    :param image_file: local file name of the image to be uploaded
    :param key: crypto key to use for encrypting the image (binary)
    :param destination: destination for the converted image (webhdfs url),
        defaults to a local proxy, passing the data to the authenticated
        connection to the Hansken remote
    :param bufsize: buffer size to use when reading image data (only applicable
        when using a local proxy as the destination)
    :param max_retries: maximum number of times to retry uploading a chunk of
        data (only applicable when using a local proxy as the destination)
    :param retry_wait: number of seconds to wait before retrying a failed chunk
        of data (only applicable when using a local proxy as the destination)
    :param auth: a `.HanskenAuthBase` object
    :param project: project to link the uploaded image to (or False)
    :param extract: whether to extract the new image (requires project to be
        set)
    :param store_key: whether to store the provided key in the key store
    :param image_tool_path: (Path to) hansken image tool in the system being used.
        For uploading from Windows this should be set to the (batch) script, e.g. 'hansken-image-tool.bat' or
        'C:\\path\\to\\tool\\hansken-image-tool.bat' (See: https://bugs.python.org/issue8224)
        Defaults to 'hansken-image-tool' which works for unix systems
    :param streaming When set to true, use the streaming upload. When set to False, use the stable chunk upload
    :param image_meta: freeform meta data for the uploaded image, description
        defaults to file base name (e.g. MyImage.E01)
    :return: a generator yielding (name, value) tuples provided by the image
        converter (see upload_info)
    """
    # implementation requires the communication marker to be whitespace
    if not upload_info.marker.isspace():
        raise ValueError('image tool communication marker ({}) needs to be whitespace'.format(upload_info.marker))

    def run_image_tool(destination, uploaded_image):
        """
        Runs hansken-image-tool, uploading an image to an HTTP endpoint.

        :param destination: an HTTP-destination to pass along to the
            ``hansken-image-tool`` command as the image destination
        :param uploaded_image: a image metadata dict to be updated in-place
            should any image metadata need to be updated
        :return: a generator yielding (name, value) tuples provided by the
            image converted (see upload_info)
        """
        command_line = [image_tool_path,
                        '--id', image_id,
                        '--format', 'NFI',
                        '--key', b64encode(key),
                        '--machine',  # hidden option request machine-readable output
                        image_file,
                        destination]

        # start image conversion, redirect stderr to stdout (text mode)
        process = Popen(command_line, universal_newlines=True, stdout=PIPE, stderr=STDOUT)  # nosec
        # limit number of lines output to 1024, but count the total number of lines captured
        console_lines = deque(maxlen=1024)
        num_lines = 0
        # process.communicate blocks, instead read from process' standard out until EOF
        for line in iter(process.stdout.readline, ''):
            # separate regular log lines from bits of info we want
            if line.startswith(upload_info.marker):
                name, value = line.strip().split(':', 1)
                if name in upload_info.names:
                    # passed info is something we need to share, find the internal name and value type
                    name, vtype = upload_info.names[name]
                    value = vtype(value)
                    yield name, value
                    if name == 'status':
                        # caught conversion status, push that to image metadata
                        uploaded_image['uploadStatus'] = value[1]
            else:
                line = line.strip()
                log.debug('log output from image tool: {}', line)
                console_lines.append(line)
                num_lines += 1

        message = 'conversion process (image id {image_id}) terminated with return code {return_code}, ' \
                  'output follows {tail}\n---\n{output}'
        tail = '' if len(console_lines) == num_lines else f'(last {len(console_lines)} of total {num_lines} lines)'
        if process.wait() != 0:
            raise ValueError(message.format(image_id=image_id if image_id is not None else 'unknown',
                                            return_code=process.returncode,
                                            tail=tail,
                                            output='\n'.join(console_lines)))

    def do_upload(connection):
        # create the image metadata of the image to be uploaded (override description if required)
        description = image_meta.pop('description', None) or path.basename(image_file)
        connection.create_image(id=image_id, description=description, **image_meta)

        # retrieve the freshly created image metadata should we need to update it
        created_image = connection.image(image_id)
        # remove the image's id, updates won't be accepted with the id in the update 'body'
        created_image.pop('id')
        # copy the created metadata so we can detect edits made during upload
        uploaded_image = dict(created_image)

        try:
            if project:
                # if requested, link image to project
                connection.link_image(project, image_id)
        except HTTPError as e:
            # not authorized to link the image (or some other problem); remove the metadata
            connection.delete_image(image_id)
            raise ValueError(f'failed to link image {image_id} to project {project}') from e

        try:
            if destination is local_proxy:
                # spin up a local WebHDFS-like server, passing data along to connection.upload_image, while passing any
                # named values generated from the image tool to the caller

                # define a wrapper around connection.upload_image to force the callback to be authenticated
                # an upload_image call potentially needs a data source that can't be reused, so make sure that any
                # potential forced re-auth happens on a repeatable call
                def force_auth_upload_callback(*args, **kwargs):
                    # call the identity_uid function to avoid the cached identity, make the remote tell us who we are
                    connection.identity_uid()
                    # now call the actual callback we intended to call before
                    return connection.upload_image(*args, **kwargs)

                with WebHDFSServer(force_auth_upload_callback,
                                   bufsize=bufsize, max_retries=max_retries, retry_wait=retry_wait,
                                   streaming=streaming) as server:
                    yield from run_image_tool('http://localhost:{}/'.format(server.server_port), uploaded_image)
            else:
                # pass any named values generated from the image tool to the caller (image tool uploads the data itself)
                yield from run_image_tool(destination, uploaded_image)
        except (KeyboardInterrupt, BrokenPipeError):
            # The user aborted the upload, or the upload has failed due to a broken pipe error
            log.warn('upload was interrupted, removing partially uploaded image {}', image_id)
            if project:
                connection.unlink_image(project, image_id)
            connection.delete_image(image_id)
            log.warn('partial upload removed')
            raise

        if uploaded_image != created_image:
            # image metadata was changed during upload, reflect changes to remote
            connection.edit_image(image_id, **uploaded_image)
        if store_key:
            connection.store_key(image_id, key)
        if project and extract:
            # if requested and there's a project available, extract traces from the uploaded image
            connection.extract(project, image_id, 'index', key)

    # call do_upload with an existing or new connection object
    if hasattr(connection_or_endpoint, 'extract'):
        # we've been handed a Connection, use it without closing it
        # NB: using isinstance here makes testing nigh impossible
        return do_upload(connection_or_endpoint)
    else:
        # using an endpoint, create local Connection that gets closed after the needed calls
        with Connection(connection_or_endpoint, auth=auth, keystore_url=keystore_url, verify=verify) as connection:
            return do_upload(connection)


def prompt_verify_key():
    """
    Prompts for a key twice, verifying the entered values are identical before
    returning it. Aborts on keyboard interrupt or EOF.

    :return: the key entered by the user
    """
    try:
        while True:
            key = getpass(prompt='Key (base64-encoded): ').strip()
            if key == getpass(prompt='Key (again): ').strip():
                return key
            else:
                print("Provided keys don't match, please try again (^C or {eof} to abort)".format(
                    eof='^Z' if 'win32' in sys.platform else '^D'
                ))
    except (KeyboardInterrupt, EOFError):  # user pressed either ^C or ^D (^Z on windows)
        raise SystemExit('Aborted')


# define command upload
upload_parser = add_command('upload',
                            run_upload_command,
                            help='upload an image to Hansken image store')
# add arguments for sub-command upload
upload_parser.add_argument('images', metavar='FILE', nargs='+', help='local file(s) to convert and upload')
upload_parser.add_argument('-o', '--output', dest='hdfs_url', metavar='ENDPOINT', default=local_proxy,
                           help=SUPPRESS)
upload_parser.add_argument('-b', '--bufsize', metavar='SIZE', type=parse_byte_size, default=8 << 20,
                           help='buffer size to be used when reading data from hansken-image-tool')
upload_parser.add_argument('-r', '--retry', metavar='N', type=int, default=8,
                           help='maximum number of times to retry uploading chunks of data '
                                '(defaults to 8, set to 0 to attempt just once, set to -1 to set no max)')
upload_parser.add_argument('-w', '--wait', metavar='SECONDS', type=float, default=10.0,
                           help='amount of time in seconds to wait between chunk upload retries')
upload_parser.add_argument('-i', '--image-id', dest='image_id', metavar='IMAGE_ID', help=SUPPRESS)
upload_parser.add_argument('-L', '--no-link', dest='link', action='store_false', default=True,
                           help='do not link uploaded image to a project (disables extraction)')
upload_parser.add_argument('-k', '--key', dest='key', metavar='KEY', default=lambda: b64encode(urandom(32)),
                           help='key to access IMAGE (base64-encoded). by default, a random key will be generated')
upload_parser.add_argument('-K', '--prompt-key', dest='key', action='store_const',
                           const=prompt_verify_key, help='prompt for a key (base64-encoded)')
upload_parser.add_argument('--streaming', action='store_true', default=False,
                           help='Stream the image data to Hansken instead of using chunked upload')
upload_parser.add_argument('-x', '--extract', action='store_true',
                           help='start extraction of traces after image upload')
upload_parser.add_argument('-s', '--store-key', dest='store_key', action='store_true', default=True,
                           help='store provided or generated key in keystore')
upload_parser.add_argument('-S', '--no-store-key', dest='store_key', action='store_false',
                           help='do not store key in keystore')
upload_parser.add_argument('--hansken-image-tool', dest='image_tool_path', default='hansken-image-tool',
                           help='(Path to) script calling the Hansken image tool. Should include script\'s'
                                'extension in Windows')



# encoding=utf-8

"""
Copyright 2015 Netherlands Forensic Institute

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from pkg_resources import resource_string


try:
    # read the resource VERSION as binary data to determine value for __version__
    __version__ = resource_string(__name__, 'VERSION').decode('utf-8').strip()
except Exception:
    __version__ = '0.0+unknown'


class envvars:  # noqa: N801
    """
    `hansken.envvars` contains the names of all :ref:`environment variables
    <command_line>` used by the commands available from the command line and
    functions like `auth.from_env <hansken.auth.from_env>` to read
    configuration values from environment variables.
    """
    pass


envvars = envvars()
for key, envvar in {'endpoint': 'HANSKEN_ENDPOINT',
                    'project': 'HANSKEN_PROJECT',
                    'keystore': 'HANSKEN_KEYSTORE_ENDPOINT',
                    'preference': 'HANSKEN_PREFERENCE_ENDPOINT',
                    'log': 'HANSKEN_LOG',
                    'idp': 'HANSKEN_IDP',
                    'idp_url': 'HANSKEN_IDP_ENDPOINT',
                    'idp_realm': 'HANSKEN_IDP_REALM',
                    'sso_url': 'HANSKEN_SSO_ENDPOINT',
                    'hdfs': 'HANSKEN_HDFS_ENDPOINT'}.items():
    setattr(envvars, key, envvar)


class fetch:  # noqa: N801
    """
    `hansken.fetch` is a sentinel value that is used to indicate that a
    resource should be automatically fetched, if possible. Resources mostly
    concern keys used to access data or metadata. `key` parameters for methods
    like `Trace.open <hansken.trace.Trace.open>` will have this value as the
    default, allowing you to manually specify a value if needed.
    """
    def __repr__(self):
        return '<auto-fetch>'  # help() shows this when fetch occurs as a default for an argument

    __str__ = __repr__


fetch = fetch()



# encoding=utf-8

from argparse import SUPPRESS
import warnings

from logbook import Logger

from hansken import __version__
from hansken.remote import Connection, ProjectContext
from hansken.tool import add_command


log = Logger(__name__)


try:
    from IPython import embed

    def shell(endpoint=None, project=None, keystore=None, preference=None, auth=None,
              verify=True, admin=False,
              **shell_locals):
        """
        Spawns an interactive shell. If an endpoint and project id are provided
        through args, a `ProjectContext <hansken.remote.ProjectContext>` instance
        is created attached to project at endpoint and exposed as local variable
        ``context``.

        :param endpoint: HTTP endpoint to a Hansken gatekeeper
        :param project: project identifier to create a context for
        :param keystore: HTTP endpoint to a Hansken keystore
        :param preference: HTTP endpoint to Hansken preference service
        :param auth: `HanskenAuthBase <hansken.auth.HanskenAuthBase>` instance to
            handle authentication, or `None`
        :param shell_locals: keyword arguments to expose as locals in the spawned
            shell
        """
        # construct multiline header
        header = ['Hansken interactive shell (client API version {})'.format(__version__)]

        if endpoint and 'connection' not in shell_locals:
            # provide a connection if not already provided as kwarg
            connection = Connection(endpoint, keystore, preference, auth=auth, verify=verify)
            shell_locals['connection'] = connection
        if admin:
            warnings.warn('use of admin in a shell has been deprecated, use connection',
                          DeprecationWarning)
        if project and 'connection' in shell_locals and 'context' not in shell_locals:
            # provide a project context if not already provided as kwarg (always created from pre-created Connection)
            context = ProjectContext(shell_locals['connection'], project)
            shell_locals['context'] = context

        # determine column width for shell locals dump
        maxlen = max(len(key) for key in shell_locals) if shell_locals else 0
        # add header lines for all shell locals, sorted by name
        header.extend('{name:<{len}} -> {value}'.format(name=name, len=maxlen, value=value)
                      for name, value in sorted(shell_locals.items()))
        # included to end banner with an empty line, like ipython itself
        header.append('')
        # start interactive session
        embed(banner2='\n'.join(header),  # print our info after the default ipython banner (including python version)
              user_ns=shell_locals,    # expose explicit locals created in shell_locals
              colors='neutral')  # work around an ipython colors bug (https://github.com/ipython/ipython/issues/11523)

    ipython_available = True
except ImportError as e:
    log.debug('shell command not available', e)
    ipython_available = False

    def shell(*args, **kwargs):
        shell_parser.error("command requires ipython, install hansken.py with the 'shell' extra or "
                           "install package 'ipython' manually")


help_text = 'spawn an interactive python shell'
if not ipython_available:
    help_text += ' (requires ipython to be installed)'


# define a command shell (ignore the return value, shell doesn't take additional arguments)
shell_parser = add_command('shell',
                           lambda args: shell(endpoint=args.endpoint,
                                              project=args.project,
                                              keystore=args.keystore,
                                              preference=args.preference,
                                              auth=args.auth,
                                              verify=args.verify,
                                              admin=args.admin),
                           help=help_text)
shell_parser.add_argument('-a', '--admin', dest='admin', action='store_true', default=False,
                          help=SUPPRESS)



# encoding=utf-8

from datetime import datetime, timedelta, timezone
from getpass import getpass
import math
from os import getcwd, makedirs, path, urandom
import sys
import threading
import time

import iso8601
from logbook import Logger

from hansken.remote import Connection, ProjectContext
from hansken.tool import add_command
from hansken.util import b64decode, b64encode

log = Logger(__name__)


def run_backup_command(args):
    """
    Starts the backup of a project. Requires a keystore
    to work for encrypted images; keys will be fetched automatically.

    :param args: an `argparse.Namespace` object carrying command line arguments
    """
    # check that we'll have an endpoint and a project (error message emulated to look like argparse's own messages)
    if not args.endpoint:
        backup_parser.error('the following argument is required: --endpoint ENDPOINT')
    if not args.project and not args.backup_modified:
        backup_parser.error('the following arguments are required: PROJECT or BACKUP_MODIFIED')
    if not args.keystore:
        backup_parser.error('the following argument is required: --keystore ENDPOINT')
    if not args.preference:
        backup_parser.error('the following argument is required: --preference ENDPOINT')
    if not args.user_backup_key:
        backup_parser.error('the following argument is required: USER_BACKUP_KEY')

    user_backup_key = args.user_backup_key

    with Connection(args.endpoint, keystore_url=args.keystore, preference_url=args.preference, auth=args.auth,
                    verify=args.verify) as connection:
        if args.backup_modified:
            projects = [project.get('id') for project in get_projects_to_backup(connection)]
            if not projects:
                log.info('No projects found to backup')
        else:
            projects = [args.project]

        keys_folder = None
        for project in projects:
            if callable(user_backup_key):
                # generate a new key for every backup if a generating lambda is provided
                user_backup_key = user_backup_key()
                keys_folder = args.keys_folder if args.keys_folder else args.download_folder

            try:
                backup(connection,
                       b64decode(user_backup_key, validate=True),
                       project=project,
                       download_backup=args.download_backup,
                       download_folder=args.download_folder,
                       keys_folder=keys_folder)
            # catch all exceptions, so backup is still performed for other projects in case of an error
            except Exception:
                log.exception('Error backing up project: {}'.format(project))


def store_user_backup_key(user_backup_key, keys_folder, project, task_id):
    """
    Stores the backup key in the keys folder
    :param user_backup_key: the key in binary (`bytes`) to write
    :param keys_folder: folder to write the user key file to
    :param project: the id of the project of the key
    :param task_id: the id of the backup task
    """
    project_path = path.join(keys_folder, project)
    makedirs(project_path, exist_ok=True)
    key_file_path = path.join(project_path, f'{task_id}.backup.key')
    with open(key_file_path, 'w') as f:
        f.write(b64encode(user_backup_key))


def is_backup_task_of_current_user(task, connection):
    """
    Checks if a task is a backup task of the current user..

    :param task: project id of project
    :param connection: instance of `.Connection`
    :return: a boolean that indicates if the project should be backed up
    """
    task_details = task.get('task')
    return task_details.get('type') == 'backup' and task_details.get('queuedBy') == connection.identity


def check_backup_needed(project, connection):
    """
    Returns if the project should be backed-up:
    This is the case when there are no backup tasks started by the current user since the last modification date of the
    project. If there is no last modification date no backup is needed.

    :param project: project as a dict
    :param connection: instance of `.Connection`
    :return: a boolean that indicates if the project should be backed up
    """
    project_id = project.get('id')
    last_modified_date = project.get('lastModifiedDate')
    if last_modified_date is not None:
        parsed_last_modified_date = iso8601.parse_date(last_modified_date)
        now = datetime.now(timezone.utc)
        open_tasks = connection.tasks('open', project_id, start=parsed_last_modified_date, end=now)

        # the earliest retrieval date for closed tasks is 8 weeks
        earliest_task_retrieval_date = now - timedelta(days=56)
        task_retrieval_date = max(earliest_task_retrieval_date, parsed_last_modified_date)
        closed_tasks = [task for task in connection.tasks('closed', project_id, start=task_retrieval_date, end=now)
                        if task.get('task', {}).get('status') == 'completed']
        all_tasks = open_tasks + closed_tasks

        backup_tasks = [task for task in all_tasks if is_backup_task_of_current_user(task, connection)]
        log.info('{} active backup tasks found for current user for project: {}', len(backup_tasks), project_id)
        return not backup_tasks
    log.info('project {} does not have a lastModificationDate set', project_id)
    return False


def get_projects_to_backup(connection):
    """
    Retrieves the projects to backup.

    :param connection: an instance of `.Connection`
    :return: list of projects to backup
    """
    return [project for project in connection.projects() if check_backup_needed(project, connection)]


def backup(connection, user_backup_key, project=False,
           download_backup=False, download_folder=getcwd(), keys_folder=None):
    """
    Starts a backup for the given project. The image keys for the project are
    retrieved from the keystore. A backup_key has to be provided to encrypt the backup.

    :param connection: an instance of `.Connection`
    :param user_backup_key: crypto key in binary (`bytes`) for encrypting the backup
    :param project: project to backup
    :param download_backup: boolean denoting if the backup should be downloaded
    :param download_folder: folder to write the backup file to
    :param keys_folder: folder to write the backup key to
    :return: task_id of the created backup task
    """
    if download_backup and not path.isdir(download_folder):
        raise ValueError('download folder "{}" does not exist or is not a folder'.format(download_folder))

    with ProjectContext(connection, project) as context:
        task_id = context.connection.backup_project(project, user_backup_key)

        if keys_folder is not None:
            store_user_backup_key(user_backup_key, keys_folder, project, task_id)

        if download_backup:
            task_finished = threading.Event()
            backup_progress = {'stage': '', 'progress': 0.0, 'status': ''}

            log.info('polling task with id: {}', task_id)

            thread = threading.Thread(target=wait_for_task,
                                      args=(context, task_id, task_finished, backup_progress))
            thread.start()

            # wait here for the result to be available before continuing
            while not task_finished.wait(timeout=5):
                print('\r{}% done...'.format(math.floor(100 * backup_progress['progress'])), end='', flush=True)
            print('\r100% done...')

            if backup_progress['status'] == 'completed':
                project_path = path.join(download_folder, project)
                makedirs(project_path, exist_ok=True)
                backup_file_path = path.join(project_path, task_id + ".backup")
                do_download_backup(context, task_id, backup_file_path)
            else:
                raise ValueError('backup task not completed successfully: {}'.format(backup_progress['status']))

        return task_id


def wait_for_task(context, task_id, task_finished, backup_progress):
    try:
        while True:
            task = context.connection.task(task_id)
            backup_task = task['task']
            backup_progress['stage'] = backup_task.get('progressStage', '')
            backup_progress['progress'] = backup_task.get('progressCompleted', 0.0)
            backup_progress['status'] = task_status = backup_task.get('status', '')

            # Possible statuses are: queued, postponed, running, cancelled, failed, completed
            if task_status not in {'queued', 'postponed', 'running'}:
                log.info('task finished with status: {}', task_status)
                task_finished.set()
                break

            time.sleep(1)
    except Exception:
        log.exception('Error waiting for backup task to complete')
        task_finished.set()


def do_download_backup(context, task_id, backup_file_path):
    log.info('downloading backup as {}', backup_file_path)
    context.connection.download_backup(task_id, backup_file_path)
    log.info('download successful.')


def prompt_verify_key():
    """
    Prompts for a key twice, verifying the entered values are identical before
    returning it. Aborts on keyboard interrupt or EOF.

    :return: the key entered by the user
    """
    try:
        while True:
            key = getpass(prompt='Key (base64-encoded): ').strip()
            if key == getpass(prompt='Key (again): ').strip():
                return key
            else:
                print("Provided keys don't match, please try again (^C or {eof} to abort)".format(
                    eof='^Z' if 'win32' in sys.platform else '^D'
                ))
    except (KeyboardInterrupt, EOFError):  # user pressed either ^C or ^D (^Z on windows)
        raise SystemExit('Aborted')


backup_parser = add_command('backup', run_backup_command, optional_project=True,
                            help='start backup of a project')
backup_parser.add_argument('-k', '--user-backup-key', dest='user_backup_key', metavar='USER_BACKUP_KEY',
                           default=lambda: b64encode(urandom(32)),
                           help='user provided key to encrypt backup (base64-encoded). '
                                'By default a random key will be generated')
backup_parser.add_argument('-K', '--prompt-user-backup-key', dest='user_backup_key', action='store_const',
                           const=prompt_verify_key, help='prompt for a user backup key (base64-encoded)')
backup_parser.add_argument('-d', '--download-backup', dest='download_backup', action='store_true',
                           help='enable backup download.')
backup_parser.add_argument('-f', '--download-folder', dest='download_folder',
                           help='folder to store the downloaded backups', default=getcwd())
backup_parser.add_argument('-m', '--backup-modified', dest='backup_modified', action='store_true',
                           help='backup projects that have been modified since the last backup', default=False)
backup_parser.add_argument('-kf', '--keys-folder', dest='keys_folder',
                           help='folder to store the generated keys', default=None)
